#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import pytest
from mlinfra.terraform.terraform import Terraform


class Test:
    # Generates a list of modules to be applied via the -target function of terraform
    def test_generate_modules_list(self, mocker):
        # Mock the necessary dependencies
        mocker.patch("os.listdir", return_value=["module1.tf.json"])
        mocker.patch(
            "builtins.open",
            mocker.mock_open(
                read_data='{"module": {"vpc": {"source": "./modules/cloud/aws/vpc","name": "aws-lakefs-k8s-vpc"}}}'
            ),
        )

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke the method
        modules_list = terraform.generate_modules_list()

        # Assert the result
        assert modules_list == ["module.vpc"]

    # Returns a string of terraform targets for the plan command
    def test_plan_output(self, mocker):
        # Mock the necessary dependencies
        mocker.patch(
            "mlinfra.terraform.terraform.Terraform.generate_terraform_config",
            return_value=("state_name", "aws_region"),
        )
        mocker.patch("mlinfra.terraform.terraform.Terraform.check_cloud_credentials")
        mocker.patch("mlinfra.terraform.terraform.Terraform.check_region_has_three_azs")
        mocker.patch("mlinfra.terraform.terraform.Terraform.check_terraform_state_storage")
        mocker.patch(
            "mlinfra.terraform.terraform.Terraform.generate_modules_list",
            return_value=["module.module1", "module.module2"],
        )

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke the method
        plan_output = terraform.plan()

        # Assert the result
        assert plan_output == " -target=module.module1 -target=module.module2"

    # Runs preliminary checks before generating the terraform config file
    def test_generate_terraform_config(self, mocker):
        # Mock the necessary dependencies
        mocker.patch("mlinfra.terraform.terraform.Terraform.check_terraform_installed")
        mocker.patch("mlinfra.terraform.terraform.Terraform.check_config_file_exists")
        mocker.patch("mlinfra.terraform.terraform.Terraform.clean_mlops_infra_folder")
        mocker.patch(
            "mlinfra.terraform.terraform.Terraform.process_config_file",
            return_value=("state_name", "aws_region"),
        )

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke the method
        state_name, aws_region = terraform.generate_terraform_config()

        # Assert the result
        assert state_name == "state_name"
        assert aws_region == "aws_region"

    # The file specified in stack_config_path does not exist
    def test_check_config_file_exists_file_not_found(self, mocker):
        # Mock the necessary dependencies
        mocker.patch("os.path.isfile", return_value=False)

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke and assert the exception
        with pytest.raises(FileNotFoundError):
            terraform.check_config_file_exists()

    # The stack config file is not a valid yaml file
    def test_check_config_file_exists_invalid_yaml(self, mocker):
        # Mock the necessary dependencies
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch("builtins.open", mocker.mock_open(read_data="invalid_yaml"))

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke and assert the exception
        with pytest.raises(ValueError):
            terraform.check_config_file_exists()

    # The stack config file is missing required keys
    def test_check_config_file_exists_missing_keys(self, mocker):
        # Mock the necessary dependencies
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch("builtins.open", mocker.mock_open(read_data='{"name": "stack_name"}\n'))

        # Initialize the class object
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)

        # Invoke and assert the exception
        with pytest.raises(ValueError):
            terraform.check_config_file_exists()

    # The test checks if the method 'generate_terraform_config' returns a tuple of strings representing the state name and AWS region
    def test_returns_state_name_and_aws_region_with_mocked_yaml_safe_load_with_mocked_configure_provider(
        self, mocker
    ):
        mocker.patch.object(Terraform, "check_terraform_installed")
        mocker.patch.object(Terraform, "check_config_file_exists")
        mocker.patch.object(Terraform, "clean_mlops_infra_folder")
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch.object(
            Terraform, "process_config_file", return_value=("stack_name", "aws_region")
        )
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)
        state_name, aws_region = terraform.generate_terraform_config()
        assert state_name == "stack_name"
        assert aws_region == "aws_region"

    # check_terraform_installed returns an error message when an exception is encountered
    def test_check_terraform_installed_returns_error_message(self, mocker):
        mocker.patch("subprocess.check_output", side_effect=Exception("An error occurred"))
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)
        result = terraform.check_terraform_installed()
        assert isinstance(result, str)
        assert "An error occurred while checking the Terraform version" in result

    # stack config file does not exist, raises FileNotFoundError
    def test_stack_config_file_not_found(self, mocker):
        mocker.patch("os.path.isfile", return_value=False)
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)
        with pytest.raises(FileNotFoundError):
            terraform.check_config_file_exists()

    # stack config file is invalid YAML, raises ValueError
    def test_stack_config_file_invalid_yaml(self, mocker):
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch("builtins.open", mocker.mock_open(read_data="invalid_yaml"))
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)
        with pytest.raises(ValueError):
            terraform.check_config_file_exists()

    # stack config file is missing required keys, raises ValueError
    def test_stack_config_file_missing_keys(self, mocker):
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch("builtins.open", mocker.mock_open(read_data='{"name": "stack_name"}\n'))
        stack_config_path = "path/to/stack/config.yaml"
        terraform = Terraform(stack_config_path)
        with pytest.raises(ValueError):
            terraform.check_config_file_exists()



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import json
from importlib import resources

import yaml
from mlinfra import modules
from mlinfra.enums.cloud_provider import CloudProvider
from mlinfra.stack_processor.deployment_processor.deployment import (
    AbstractDeployment,
)
from mlinfra.utils.utils import generate_tf_json


class KubernetesDeployment(AbstractDeployment):
    def __init__(
        self,
        stack_name: str,
        provider: CloudProvider,
        region: str,
        deployment_config: yaml,
    ):
        """
        Initialize a new instance of the KubernetesDeployment class.

        Parameters:
        - stack_name (str): The name of the stack.
        - provider (CloudProvider): The cloud provider for the deployment.
        - region (str): The region for the deployment.
        - deployment_config (yaml): The deployment configuration.
        """
        super(KubernetesDeployment, self).__init__(
            stack_name=stack_name,
            provider=provider,
            region=region,
            deployment_config=deployment_config,
        )

    def generate_required_provider_config(self):
        """
        Generate the required provider configuration for the Kubernetes deployment.

        This method reads the necessary provider configuration files and generates the required provider configuration for the Kubernetes deployment.
        It updates the 'terraform.tf.json' file with the required provider information.

        Parameters:
        - None

        Returns:
        - None
        """
        with open(
            resources.files(modules) / f"cloud/{self.provider.value}/terraform.tf.json",
            "r",
        ) as data_json:
            data = json.load(data_json)

            # TODO: Update token generation for all providers
            required_providers = ["random", "kubernetes", "helm"]

            for required_provider in required_providers:
                with open(
                    resources.files(modules)
                    / f"terraform_providers/{required_provider}/terraform.tf.json",
                    "r",
                ) as provider_tf:
                    provider_tf_json = json.load(provider_tf)
                data["terraform"]["required_providers"].update(
                    provider_tf_json["terraform"]["required_providers"]
                )

            data["terraform"].update(self.get_provider_backend(provider=self.provider))

            generate_tf_json(module_name="terraform", json_module=data)

    def generate_k8s_helm_provider_config(self):
        data = {"provider": {}}

        # TODO: Update token generation for all providers
        providers = ["kubernetes", "helm"]

        for provider in providers:
            with open(
                resources.files(modules) / f"terraform_providers/{provider}/provider.tf.json",
                "r",
            ) as provider_tf:
                provider_tf_json = json.load(provider_tf)
            data["provider"].update(provider_tf_json["provider"])

        generate_tf_json(module_name="k8s_provider", json_module=data)

    def generate_deployment_config(self):
        """
        Generate the deployment configuration for the Kubernetes deployment.

        This method generates the deployment configuration for the Kubernetes deployment based on the specified provider.
        It injects the necessary modules and configurations into the Terraform configuration files.

        Parameters:
        - None

        Returns:
        - None

        Raises:
        - ValueError: If the specified provider is not supported.
        """
        if self.provider == CloudProvider.AWS:
            # TODO: Make these blocks generic
            # inject vpc module
            vpc_json_module = {"module": {"vpc": {}}}
            vpc_json_module["module"]["vpc"]["source"] = "./modules/cloud/aws/vpc"

            vpc_json_module["module"]["vpc"]["name"] = f"{self.stack_name}-vpc"

            if "config" in self.deployment_config and "vpc" in self.deployment_config["config"]:
                for vpc_config in self.deployment_config["config"]["vpc"]:
                    vpc_json_module["module"]["vpc"][vpc_config] = self.deployment_config["config"][
                        "vpc"
                    ].get(vpc_config, None)

            generate_tf_json(module_name="vpc", json_module=vpc_json_module)
            # inject vpc module

            # inject k8s module
            k8s_json_module = {"module": {"eks": {}}}
            k8s_json_module["module"]["eks"]["source"] = "./modules/cloud/aws/eks/tf_module"
            k8s_json_module["module"]["eks"]["cluster_name"] = f"{self.stack_name}-cluster"
            k8s_json_module["module"]["eks"]["vpc_id"] = "${ module.vpc.vpc_id }"
            k8s_json_module["module"]["eks"]["subnet_ids"] = "${ module.vpc.private_subnets_ids }"

            if (
                "config" in self.deployment_config
                and "kubernetes" in self.deployment_config["config"]
            ):
                # read values from the yaml config file
                with open(
                    resources.files(modules) / f"cloud/{self.provider.value}/eks/eks.yaml",
                    "r",
                    encoding="utf-8",
                ) as tf_config:
                    eks_application_config = yaml.safe_load(tf_config.read())

                # check if values exist in the yaml config file
                if (
                    eks_application_config is not None
                    and "inputs" in eks_application_config
                    and eks_application_config["inputs"]
                ):
                    # iterate through the config defined in the deployment section
                    # of the stack file
                    for k8s_config in self.deployment_config["config"]["kubernetes"]:
                        # if the application config exists in the eks config and is
                        # configured to be user_facing, only then add this config
                        config_lookup = next(
                            (
                                item
                                for item in eks_application_config["inputs"]
                                if item["name"] == k8s_config and item["user_facing"]
                            ),
                            None,
                        )
                        if config_lookup is not None:
                            k8s_json_module["module"]["eks"][k8s_config] = self.deployment_config[
                                "config"
                            ]["kubernetes"].get(k8s_config, None)
                        else:
                            print(
                                """
                                WARNING: The config value {k8s_config} is not user facing.
                                Please check the eks.yaml config file to see if this is a valid config value.
                                """
                            )

            generate_tf_json(module_name="eks", json_module=k8s_json_module)

            # TODO: read defaults from the config file if anything is missing

            if (
                "config" in self.deployment_config
                and "node_groups" in self.deployment_config["config"]
            ):
                # inject k8s module
                nodegroups_json_module = {"module": {"eks_nodegroup": {"node_groups": []}}}
                nodegroups_json_module["module"]["eks_nodegroup"][
                    "source"
                ] = "./modules/cloud/aws/eks_nodegroup/tf_module"

                for nodegroup_config in self.deployment_config["config"]["node_groups"]:
                    nodegroup_object = {}
                    nodegroup_object = nodegroup_config
                    nodegroup_object["subnet_ids"] = "${ module.vpc.private_subnets_ids }"
                    nodegroup_object["cluster_name"] = "${ module.eks.cluster_name }"
                    nodegroup_object["cluster_version"] = "${ module.eks.cluster_version }"
                    nodegroup_object[
                        "cluster_primary_security_group_id"
                    ] = "${ module.eks.cluster_primary_security_group_id }"
                    nodegroup_object[
                        "node_security_group_id"
                    ] = "${ module.eks.node_security_group_id }"

                    nodegroups_json_module["module"]["eks_nodegroup"]["node_groups"].append(
                        nodegroup_object
                    )

                generate_tf_json(module_name="nodegroups", json_module=nodegroups_json_module)

        elif self.provider == CloudProvider.GCP:
            raise ValueError(f"Provider {self.provider} is not yet supported")
        elif self.provider == CloudProvider.AZURE:
            raise ValueError(f"Provider {self.provider} is not yet supported")
        else:
            raise ValueError(f"Provider {self.provider} is not supported")

    def configure_deployment(self):
        self.generate_required_provider_config()
        self.generate_k8s_helm_provider_config()
        self.generate_deployment_config()


if __name__ == "__main__":
    pass



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import tasks
from invoke import Collection, Program
from mlinfra.utils.constants import VERSION

namespace = Collection()
namespace.add_task(tasks.terraform)
namespace.add_task(tasks.estimate_cost)
namespace.add_task(tasks.generate_terraform_config)

program = Program(version=VERSION, namespace=namespace, name="mlinfra")


def cli() -> None:
    """Welcome to mlinfra cli! Deploy any MLOps tooling at the click of a button.

    Github: https://github.com/mlinfra-io/mlinfra
    Documentation: https://mlinfra.io/
    """
    program.run()


if __name__ == "__main__":
    program.cli()



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from importlib.metadata import version
from time import time
from typing import Final

# TODO: update this to path.cwd
TF_PATH = "./.mlops_stack"

CREATE_VPC_DB_SUBNET = ["remote_tracking"]

DEV_VERSION: Final = "dev"
MLINFRA_DISABLE_REPORTING: Final = "MLINFRA_DISABLE_REPORTING"
SESSION_ID: Final = int(time() * 1000)

VERSION: Final = version("mlinfra")

EXECUTABLE_PATH = "/usr/local/bin"



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import pytest
from mlinfra.stack_processor.provider_processor.provider import AbstractProvider


class TestAbstractProvider:
    # AbstractProvider object can be initialized with stack name and configuration object.
    def test_initialize_with_stack_name_and_config(self):
        stack_name = "test_stack"
        config = {"key": "value"}

        class ConcreteProvider(AbstractProvider):
            def configure_provider(self):
                pass

        provider = ConcreteProvider(stack_name, config)

        assert provider.stack_name == stack_name
        assert provider.config == config

    # Subclasses can implement configure_provider method to configure the provider.
    def test_configure_provider_method(self):
        class TestProvider(AbstractProvider):
            def configure_provider(self):
                return "Provider configured"

        provider = TestProvider("test_stack", {"key": "value"})

        assert provider.configure_provider() == "Provider configured"

    # stack_name parameter is not provided during initialization.
    def test_missing_stack_name_parameter(self):
        with pytest.raises(TypeError):
            AbstractProvider(config={"key": "value"})

    # config parameter is not provided during initialization.
    def test_missing_config_parameter(self):
        with pytest.raises(TypeError):
            AbstractProvider(stack_name="test_stack")

    # Subclasses can override __init__ method to add additional parameters.
    def test_subclass_override_init_method(self):
        class TestProvider(AbstractProvider):
            def __init__(self, stack_name, config, additional_param):
                super().__init__(stack_name, config)
                self.additional_param = additional_param

            def configure_provider(self):
                pass

        provider = TestProvider("test_stack", {"key": "value"}, "additional")

        assert provider.stack_name == "test_stack"
        assert provider.config == {"key": "value"}
        assert provider.additional_param == "additional"



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from mlinfra.stack_processor.provider_processor.aws_provider import AWSProvider


class TestAWSProvider:
    # AWSProvider can be instantiated with a stack name and a configuration object
    def test_instantiation_with_stack_name_and_config(self):
        stack_name = "test_stack"
        config = {"account_id": "123456789", "region": "us-west-2"}
        provider = AWSProvider(stack_name, config)

        assert provider.stack_name == stack_name
        assert provider.config == config
        assert provider.account_id == config["account_id"]
        assert provider.region == config["region"]
        assert provider.access_key is None
        assert provider.secret_key is None
        assert provider.role_arn is None

    # AWSProvider can retrieve the name of the statefile for a given stack and region
    def test_get_statefile_name(self):
        stack_name = "test_stack"
        config = {"account_id": "123456789", "region": "us-west-2"}
        provider = AWSProvider(stack_name, config)

        expected_statefile_name = f"tfstate-{stack_name}-{config['region']}"

        assert provider.get_statefile_name() == expected_statefile_name

    # AWSProvider can be instantiated with a configuration object that does not contain an access key, a secret key, or a role ARN
    def test_instantiation_without_access_key_secret_key_role_arn(self):
        stack_name = "test_stack"
        config = {"account_id": "123456789", "region": "us-west-2"}
        provider = AWSProvider(stack_name, config)

        assert provider.stack_name == stack_name
        assert provider.config == config
        assert provider.account_id == config["account_id"]
        assert provider.region == config["region"]
        assert provider.access_key is None
        assert provider.secret_key is None
        assert provider.role_arn is None

    # AWSProvider can be instantiated with a configuration object that does not contain an account ID or a region
    def test_instantiation_without_account_id_region(self):
        stack_name = "test_stack"
        config = {"access_key": "access_key", "secret_key": "secret_key", "role_arn": "role_arn"}
        provider = AWSProvider(stack_name, config)

        assert provider.stack_name == stack_name
        assert provider.config == config
        assert provider.account_id is None
        assert provider.region is None
        assert provider.access_key == config["access_key"]
        assert provider.secret_key == config["secret_key"]
        assert provider.role_arn == config["role_arn"]



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

import time

from boto3 import client
from botocore.config import Config
from botocore.exceptions import ClientError


class StateHelper:
    """
    The StateHelper class manages the creation and configuration of
    AWS S3 buckets and DynamoDB tables for storing Terraform state.
    """

    def __init__(self, state: str, region: str) -> None:
        """
        Gets provider json and initializes provider variables
        """
        self.bucket_name = state
        self.dynamodb_table = state
        self.region = region

    def manage_aws_state_storage(self) -> None:
        """
        Creates aws state storage resources (S3 bucket and DynamoDB table)
        if they do not exist.
        """
        s3 = client("s3", config=Config(region_name=self.region))
        dynamodb = client("dynamodb", config=Config(region_name=self.region))

        # checks if bucket exists
        try:
            s3.head_bucket(
                Bucket=self.bucket_name,
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "AuthFailure":
                raise Exception(
                    "The AWS Credentials are not configured properly.\n"
                    "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html#configuration"
                    "for more information."
                )
            elif e.response["Error"]["Code"] == "AccessDenied":
                raise Exception(
                    f"We were unable to access the S3 bucket, {self.bucket_name} on your AWS account.\n"
                    "Possible Issues: \n"
                    " - Bucket name is not unique and might be present in some other Account. \n"
                    "Try updating the name in Configuration file to something else.\n"
                    " - It could also mean that your AWS account has insufficient permissions.\n"
                    "Please fix these issues and try again!"
                )
            elif e.response["Error"]["Code"] == "404":
                print("S3 bucket for terraform state not found, creating a new one")
                self._setup_bucket(
                    s3_client=s3,
                    region=self.region,
                    bucket_name=self.bucket_name,
                    bucket_exists=False,
                )
            else:
                raise Exception(
                    f"{e.response['Error']['Code']} error with the message "
                    f"{e.response['Error']['Message']}"
                )

        # sets bucket versioning and lifecycle rules
        self._setup_bucket(
            s3_client=s3,
            region=self.region,
            bucket_name=self.bucket_name,
            bucket_exists=True,
        )

        # describe dynamodb table
        # create if error occurs in table describe
        try:
            dynamodb.describe_table(TableName=self.dynamodb_table)
        except ClientError as e:
            if e.response["Error"]["Code"] != "ResourceNotFoundException":
                raise Exception(
                    "When trying to determine the status of the state dynamodb table, we got an "
                    f"{e.response['Error']['Code']} error with the message "
                    f"{e.response['Error']['Message']}"
                )
            print("Dynamodb table for terraform state not found, creating a new one")
            dynamodb.create_table(
                TableName=self.dynamodb_table,
                KeySchema=[{"AttributeName": "LockID", "KeyType": "HASH"}],
                AttributeDefinitions=[{"AttributeName": "LockID", "AttributeType": "S"}],
                BillingMode="PROVISIONED",
                ProvisionedThroughput={
                    "ReadCapacityUnits": 20,
                    "WriteCapacityUnits": 20,
                },
            )

    def _setup_bucket(self, s3_client, region: str, bucket_name: str, bucket_exists: bool = False):
        """
        Creates state bucket if it does not exist.
        Configures bucket versioning and lifecycle rules.
        """
        if not bucket_exists:
            try:
                if region == "us-east-1":
                    s3_client.create_bucket(Bucket=bucket_name)
                else:
                    s3_client.create_bucket(
                        Bucket=bucket_name,
                        CreateBucketConfiguration={"LocationConstraint": region},
                    )
                time.sleep(10)
            except ClientError as e:
                raise Exception(
                    f"When trying to create a new bucket with name {bucket_name}, we got an "
                    f"{e.response['Error']['Code']} error with the message "
                    f"{e.response['Error']['Message']}"
                )

        # we are not checking for bucket encryption
        # as all new buckets are encrypted by default since 2016

        # checking for bucket versioning
        # enable if not enabled
        try:
            response = s3_client.get_bucket_versioning(Bucket=bucket_name).get("Status")
            if response == "Enabled":
                print("Versioning is already enabled on state bucket")
            else:
                print("Versioning is not yet enabled on state bucket")
                s3_client.put_bucket_versioning(
                    Bucket=bucket_name,
                    VersioningConfiguration={"Status": "Enabled"},
                )
                print("Versioning is now enabled on state bucket")
        except ClientError as e:
            if e.response["Error"]["Code"] != "NoSuchBucket":
                raise Exception(
                    f"When trying to check if versioning is enabled on the state bucket, we got an "
                    f"{e.response['Error']['Code']} error with the message "
                    f"{e.response['Error']['Message']}"
                )
            print("State bucket does not exist")

        # checking for bucket lifecycle configuration
        # enable if it does not exist
        try:
            response = s3_client.get_bucket_lifecycle_configuration(Bucket=bucket_name).get("Rules")
            if response is not None:
                print("Bucket lifecycle configuration already exists on state bucket")
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchLifecycleConfiguration":
                print("Bucket lifecycle configuration does not exists on state bucket")
                s3_client.put_bucket_lifecycle(
                    Bucket=bucket_name,
                    LifecycleConfiguration={
                        "Rules": [
                            {
                                "ID": "default",
                                "Prefix": "/",
                                "Status": "Enabled",
                                "NoncurrentVersionTransition": {
                                    "NoncurrentDays": 30,
                                    "StorageClass": "GLACIER",
                                },
                                "NoncurrentVersionExpiration": {"NoncurrentDays": 60},
                                "AbortIncompleteMultipartUpload": {"DaysAfterInitiation": 10},
                            },
                        ]
                    },
                )
                print("Bucket lifecycle configuration now exsits on state bucket")



#     Copyright (c) mlinfra 2024. All Rights Reserved.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at:
#         https://www.apache.org/licenses/LICENSE-2.0
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#     or implied. See the License for the specific language governing
#     permissions and limitations under the License.

from invoke import task
from mlinfra.amplitude import amplitude_client
from mlinfra.terraform.terraform import Terraform
from mlinfra.utils.constants import TF_PATH


@task(
    pre=[],
    help={
        "stack_config_path": "Path of the config file",
    },
)
def generate_terraform_config(
    ctx,
    stack_config_path: str,
) -> None:
    f"""
    Generates the terraform config in the {TF_PATH} folder path
    """
    modules = Terraform(stack_config_path).plan()
    current_properties = {"modules": modules}

    amplitude_client.send_event(
        amplitude_client.START_GEN_TERRAFORM_EVENT,
        event_properties=current_properties,
    )
    ctx.run(f"terraform -chdir={TF_PATH} init")
    print(
        f"""
            Terraform config has been generated in the {TF_PATH} folder.
        """
    )
    amplitude_client.send_event(
        amplitude_client.END_GEN_TERRAFORM_EVENT,
        event_properties=current_properties,
    )


@task(
    pre=[],
    help={
        "stack_config_path": "Path of the config file",
    },
)
def estimate_cost(
    ctx,
    stack_config_path: str,
) -> None:
    """
    Estimate cost of the contents of config file
    """
    Terraform(stack_config_path).plan()

    ctx.run(f"terraform -chdir={TF_PATH} init")
    ctx.run(
        f"terraform -chdir={TF_PATH} plan -no-color -lock=false -input=false -compact-warnings -out tfplan.binary"
    )
    ctx.run(f"terraform -chdir={TF_PATH} show -no-color -json tfplan.binary > {TF_PATH}/plan.json")
    ctx.run(f"infracost diff --show-skipped --no-cache --path {TF_PATH}/plan.json")
    amplitude_client.send_event(
        amplitude_client.COST_ESTIMATOR_EVENT,
        event_properties={},
    )


@task(
    pre=[],
    help={
        "stack_config_path": "Path of the config file",
        "action": "Allowed actions are plan, destroy, apply, force-unlock (not yet available). Defaults to plan",
        "args": "Additional args to pass to terraform. Defaults to none",
    },
)
def terraform(
    ctx,
    stack_config_path: str,
    action: str = "plan",
    args: str = "",
) -> None:
    """
    Run terraform for the config file with the given action and args.
    """
    targets_list = Terraform(stack_config_path=stack_config_path).plan()
    targets_list = ""

    ctx.run(f"terraform -chdir={TF_PATH} init")

    if action == "plan":
        args += f"{targets_list} -lock=false -input=false -compact-warnings"
        amplitude_client.send_event(
            amplitude_client.PLAN_EVENT,
            event_properties={},
        )
        ctx.run(f"terraform -chdir={TF_PATH} plan {args} -out tfplan.binary")
    elif action == "apply":
        args += f"{targets_list} -lock=false -input=false -compact-warnings"
        amplitude_client.send_event(
            amplitude_client.APPLY_EVENT,
            event_properties={},
        )
        ctx.run(f"terraform -chdir={TF_PATH} plan {args} -out tfplan.binary")
        ctx.run(f"terraform -chdir={TF_PATH} apply -auto-approve tfplan.binary")
    elif action == "destroy":
        args = "-auto-approve"
        amplitude_client.send_event(
            amplitude_client.DESTROY_EVENT,
            event_properties={},
        )
        ctx.run(f"terraform -chdir={TF_PATH} destroy {args}")
    # elif action == "force-unlock":
    #     file_processor.force_unlock()
    #     action = f"plan {args} -lock=false"
